import tempfile
import gzip
import shutil
import warnings
from itertools import chain

import numpy as np

from heparchy.data.event import ShowerData
from heparchy.utils import structure_pmu


class HepMC:
    """Returns an iterator over events in the given HepMC file.
    Event data is provided as a `heparchy.data.ShowerData` object.

    Parameters
    ----------
    path : string
        Location of the HepMC file.
        If this file is compressed with gzip, a temporary decompressed
        file will be created and cleaned up within the scope of the
        context manager.

    See also
    --------
    `heparchy.data.ShowerData` : Container for Monte-Carlo shower data.

    Examples
    --------
    >>> with HepMC('test.hepmc.gz') as hep_f:
    ...     for event in hep_f:
    ...         print(event.pmu)
    [( 0.        ,  0.        ,  6.49999993e+03, 6.50000000e+03)
     ( 1.42635177, -1.2172366 ,  1.36418624e+03, 1.36418753e+03)
     ( 0.21473153, -0.31874408,  1.04370539e+01, 1.04441276e+01) ...
     (-2.40079685,  1.56211274,  2.78633756e+00, 3.99596030e+00)
     (-0.34612959,  0.37377605,  5.25064994e-01, 7.31578753e-01)
     (-0.00765114,  0.00780012, -8.58527843e-03, 1.38956379e-02)]
    [( ...

    Notes
    -----
    If you wish to keep a given event in memory, you must use
    `ShowerData`'s `copy()` method.
    This is because the `HepMC` iterator avoids the substantial cost of
    repeated object instantiation by re-using a single `ShowerData`
    instance, with its efficient setter methods to update the data it
    contains.
    """

    import pyhepmc_ng as __hepmc
    import networkx as __nx


    def __init__(self, path):
        from typicle import Types
        self.__types = Types()
        self.path = path
        self.__gunz_f = None
        self.data = ShowerData.empty()
    # context manager
    def __enter__(self):
        try:
            self.__buffer = self.__hepmc.open(self.path, 'r')
        except UnicodeDecodeError:
            self.__gunz_f = tempfile.NamedTemporaryFile()
            with gzip.open(self.path, 'rb') as gz_f:
                shutil.copyfileobj(gz_f, self.__gunz_f)
            self.__buffer = self.__hepmc.open(self.__gunz_f.name, 'r')
        except:
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.__buffer.close()
        if self.__gunz_f is not None:
            self.__gunz_f.close()

    # iterable
    def __iter__(self):
        self.__iter = iter(self.__buffer)
        return self

    def __next__(self):
        # make contents of file available everywhere
        self.__content = next(self.__iter)
        # read in the particle data
        self.data.flush_cache()
        (self.data.edges,
         self.data.pmu,
         self.data.pdg,
         self.data.final) = self.__pcl_data()
        return self.data

    def __pcl_data(self):
        pcls = self.__content.particles
        node_id = lambda obj: int(obj.id)
        def pcl_data(pcl):
            edge_idxs = [pcl.production_vertex, pcl.end_vertex]
            edge_idxs = tuple(node_id(vtx) if vtx != None
                              else node_id(pcl)
                              for vtx in edge_idxs)
            pmu, pdg, status = tuple(pcl.momentum), pcl.pid, pcl.status
            return edge_idxs, pmu, pdg, status
        edges, pmus, pdgs, statuses = zip(*map(pcl_data, pcls))
        edges = np.fromiter(chain.from_iterable(edges), dtype=self.__types.int)
        edges = edges.reshape((-1, 2))
        pmu = np.array(list(pmus), dtype=self.__types.pmu[0][1])
        pmu = structure_pmu(pmu)
        pdg = np.fromiter(pdgs, dtype=self.__types.int)
        is_leaf = np.fromiter(
                map(lambda status: status == 1, statuses),
                dtype=self.__types.bool
                )
        return edges, pmu, pdg, is_leaf



import numpy as np

import warnings
from functools import wraps
from typing import Callable, Any

import typicle
from heparchy.annotate import AnyVector, DoubleVector, IntVector


def event_key_format(evt_num: int) -> str:
    return f"event_{evt_num:09}"


def deprecated(func: Callable[..., Any]) -> Callable[..., Any]:
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """
    @wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn(f"Call to deprecated function {func.__name__}.",
                      category=DeprecationWarning,
                      stacklevel=2)
        return func(*args, **kwargs)
    return new_func


_types = typicle.Types()


def structure_pmu(array: AnyVector) -> AnyVector:
    """Helper function to convert 4 column array into structured array
    representing 4-momenta of particles.
    
    Parameters
    ----------
    array : numpy ndarray of floats, with shape (num particles, 4)
        The 4-momenta of the particles, arranged in columns.
        Columns must be in order (x, y, z, e).

    See also
    --------
    structure_pmu_components : structured array from seperate 1d arrays
                               of momentum components.

    Notes
    -----
    As the data-type of the input needs to be recast, the output is
    a copy of the original data, not a view on it. Therefore it uses
    additional memory, so later changes to the original will not
    affect the returned array, and vice versa.
    """
    if array.dtype != _types.pmu:
        struc_array = array.astype(_types.pmu[0][1])
        struc_array = struc_array.view(dtype=_types.pmu, type=np.ndarray)
        struc_pmu = struc_array.copy().squeeze()
    else:
        struc_pmu = array
    return struc_pmu


def structure_pmu_components(x: DoubleVector, y: DoubleVector, z: DoubleVector,
                             e: DoubleVector) -> AnyVector:
    """Helper function to convert 1d arrays of 4-momentum components
    for particles into a single structured array.

    Parameters
    ----------
    x, y, z, e : numpy ndarray of floats
        components of the 4-momenta

    See also
    --------
    structure_pmu : structured array from single array with 4 columns.
    """
    data = [x, y, z, e]
    shapes = set(col.shape for col in data)
    if len(shapes) > 1:
        raise ValueError('Shapes of all input arrays must be equal')
    return structure_pmu(np.vstack(data).T)


def unstructure_pmu(struc_array: AnyVector,
                    dtype=_types.pmu[0][1]) -> DoubleVector:
    """Returns view of 4-momentum data in columns (x, y, z, e) without
    as ordinary ndarray, without named columns.

    Parameters
    ----------
    struc_array : numpy ndarray with structured dtype names (x, y, z, e)
        Input structured 4-momentum array.
    dtype : data-type, optional
        Data type of named columns in array. Important if you created
        the structured array manually, rather than relying on the
        heparchy helper functions.
        Omitting results in default of '<f4' or np.int32.

    Notes
    -----
    As this returns a view object on the original data, modifying
    elements in the resulting array will also modify the original.
    """
    return struc_array.view(dtype=dtype).reshape((-1, 4))


def structure_edges(edges: IntVector) -> AnyVector:
    if edges.dtype != _types.edge:
        struc_array = edges.astype(_types.int)
        struc_array = struc_array.view(dtype=_types.edge, type=np.ndarray)
        struc_edges = struc_array.copy().squeeze()
    else:
        struc_edges = edges
    return struc_edges


def unstructure_edges(struc_array: AnyVector,
                      dtype=_types.int) -> IntVector:
    return struc_array.view(dtype=dtype).reshape((-1, 2))



"""
`heparchy.write.hdf`
===================

Provides the interface to write HEP data to the heparchy HDF5 format.
"""
from __future__ import annotations
from typing import (
        Any, Optional, Union, Sequence, Dict, Set, Iterator, Generic, TypeVar,
        Callable, Tuple, Type, List)
from collections.abc import MutableMapping, Iterable
from pathlib import Path
import warnings
from enum import Enum
from functools import partial


import numpy.typing as npt
import h5py
from h5py import Group, File

import heparchy as hrc
from heparchy.utils import deprecated, event_key_format
from .base import WriterBase, ProcessWriterBase, EventWriterBase
from heparchy.annotate import (
        IntVector, HalfIntVector, DoubleVector, BoolVector, AnyVector)


__all__: List[str] = []

_WRITE_ONLY_MSG = "Attribute is write-only."

MapGeneric = TypeVar("MapGeneric")
IterItem = TypeVar("IterItem")
WriterType = Union["HdfEventWriter", "HdfProcessWriter"]
ExportType = TypeVar("ExportType", Callable, Type)


def _export(procedure: ExportType) -> ExportType:
    __all__.append(procedure.__name__)
    return procedure


@_export
class Compression(Enum):
    """Sets the compression algorithm used to store the datasets.
    :group: hepwrite
    """
    LZF = "lzf"
    GZIP = "gzip"


@_export
class WriteOnlyError(RuntimeError):
    """Raised when trying to access write-only data.
    :group: hepwrite
    """


@_export
class OverwriteWarning(RuntimeWarning):
    """A warning to be raised when a user writes a piece of data twice.
    :group: hepwrite
    """


def _mk_dset(grp: Group, name: str, data: AnyVector, shape: tuple,
             dtype: npt.DTypeLike, compression: Compression,
             compression_level: Optional[int]) -> None:
    """Generic dataset creation and population function.
    Wrap in methods exposed to the user interface.
    """
    if name in grp:
        warnings.warn(f"Overwriting {name}", OverwriteWarning)
        del grp[name]
    # check data can be broadcasted to dataset:
    if data.squeeze().shape != shape:
        raise ValueError(
                f"Input data shape {data.shape} "
                f"incompatible with dataset shape {shape}."
                )
    kwargs: Dict[str, Any] = dict(
            name=name,
            shape=shape,
            dtype=dtype,
            shuffle=True,
            compression=compression.value,
            )
    cmprs_lvl = compression_level
    if cmprs_lvl is not None:
        kwargs["compression_opts"] = cmprs_lvl
    dset = grp.create_dataset(**kwargs)
    dset[...] = data


def _mask_setter(writer: WriterType, name: str, data: BoolVector) -> None:
    if not isinstance(writer, HdfEventWriter):
        raise ValueError("Can't set masks on processes")
    writer._set_num_pcls(data)
    _mk_dset(
            writer._mask_grp,
            name=name,
            data=data,
            shape=data.shape,
            dtype=writer._types.bool,
            compression=writer._proc._file_obj._cmprs,
            compression_level=writer._proc._file_obj._cmprs_lvl,
            )


def _custom_setter(writer: WriterType, name: str, data: AnyVector) -> None:
    if not isinstance(writer, HdfEventWriter):
        raise ValueError("Can't set custom datasets on processes")
    _mk_dset(
            writer._custom_grp,
            name=name,
            data=data,
            shape=data.shape,
            dtype=data.dtype,
            compression=writer._proc._file_obj._cmprs,
            compression_level=writer._proc._file_obj._cmprs_lvl,
            )


def _meta_setter(writer: WriterType, name: str, data: Any) -> None:
    writer._grp.attrs[name] = data


@_export
class MapWriter(Generic[MapGeneric], MutableMapping[str, MapGeneric]):
    """Provides a dictionary-like interface for writing user-named
    datasets or metadata.

    :group: hepwrite

    Parameters
    ----------
    writer : HdfProcessWriter | HdfEventWriter
        The writer instance to which the MapWriter will be attached.
    setter_func : callable
        The strategy used to write the user-named data with the writer.

    Raises
    ------
    WriteOnlyError
        Dictionary-like behaviour which requires read-access for
        datasets, as access to the data is strictly write-only.

    Notes
    -----
    This is a concrete subclass of collections.abc.MutableMapping, and
    thus includes all functionality expected of a dictionary, including
    iteration over keys, subscriptable setters and deleters of datasets,
    etc.
    """
    def __init__(self,
                 writer: WriterType,
                 setter_func: Callable[[WriterType, str, MapGeneric], None]
                 ) -> None:
        self.writer = writer
        self._names: Set[str] = set()
        self._setter_func = setter_func

    def __repr__(self) -> str:
        dset_repr = "<Write-Only Data>"
        kv = ", ".join(map(lambda k: f"\'{k}\': {dset_repr}", self._names))
        return "{" + f"{kv}" + "}"

    def __len__(self) -> int:
        return len(self._names)

    def __getitem__(self, name: str) -> MapGeneric:
        raise WriteOnlyError("Value is write-only.")

    def __setitem__(self, name: str, data: MapGeneric) -> None:
        self._setter_func(self.writer, name, data)
        self._names.add(name)

    def __delitem__(self, name: str) -> None:
        self._names.remove(name)
        if self._setter_func.__name__ == "_meta_setter":
            del self.writer._grp.attrs[name]
            return
        del self.writer._grp[name]

    def __iter__(self) -> Iterator[str]:
        for name in self._names:
            yield name

    def _flush(self) -> Tuple[str, ...]:
        data = tuple(self._names)
        self._names = set()
        return data


@_export
class HdfEventWriter(EventWriterBase):
    """Context manager interface to create and write events.

    :group: hepwrite

    Attributes
    ----------
    edges : numpy.ndarray[int]
        A structured array providing edges for a graph representation,
        with fields "in" and "out".
    edge_weights : numpy.ndarray[double]
        Weights attributed to each edge.
    pmu : numpy.ndarray[double]
        A structured array providing particle 4-momenta, with fields
        "x", "y", "z", "e"
    color : numpy.ndarray[int]
        A structured array with fields "color" and "anti-color".
    pdg : numpy.ndarray[int]
        PDG codes providing the identity of every particle in the event.
    status : numpy.ndarray[int]
        Status codes, describing the reason for the generation of each
        particle in the event.
    helicity : numpy.ndarray[int]
        Spin polarisations of every particle in the event.
    masks : MapWriter[numpy.ndarray[bool]]
        Write-only dictionary-like interface to set boolean masks
        over the particle datasets.
    custom : MapWriter[numpy.ndarray]
        Write-only dictionary-like interface to set user-defined
        custom datasets.
    custom_meta : MapWriter
        Write-only dictionary-like interface to set user-defined
        metadata on the event.

    Examples
    --------
    Store custom metadata to an event:

        >>> import heparchy
        >>>
        >>> with heparchy.write.HdfWriter("showers.hdf5") as hep_file:
        >>>     with hep_file.new_process("higgs") as proc:
        >>>         with proc.new_event() as event:
        >>>             event.custom_meta["num_jets"] = 4

    """
    __slots__ = ("_types", "__grp_obj", "_idx", "_num_pcls", "_evt", "masks",
                 "custom", "custom_meta", "_num_edges")

    def __init__(self, proc: HdfProcessWriter) -> None:
        from typicle import Types
        self._types = Types()
        self._proc = proc  # pointer to parent group obj
        self._idx = proc._evt_idx  # index for current event
        self._num_pcls: Optional[int] = None
        self._num_edges = 0
        self._grp: Group
        self._custom_grp: Group
        self._mask_grp: Group
        self.masks: MapWriter[BoolVector]
        self.custom: MapWriter[AnyVector]
        self.custom_meta: MapWriter[Any]

    def __enter__(self: HdfEventWriter) -> HdfEventWriter:
        self._grp = self._proc._grp.create_group(
                event_key_format(self._idx))
        self._custom_grp = self._grp.create_group("custom")
        self._mask_grp = self._grp.create_group("masks")
        self.masks = MapWriter(self, _mask_setter)
        self.custom = MapWriter(self, _custom_setter)
        self.custom_meta = MapWriter(self, _meta_setter)
        self._mk_dset = partial(
                _mk_dset,
                grp=self._grp,
                compression=self._proc._file_obj._cmprs,
                compression_level=self._proc._file_obj._cmprs_lvl,
                )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        # increasing the index for the group to create next event
        # also acts as running counter for number events
        if self._num_pcls is None:
            self._num_pcls = 0
        self._grp.attrs["num_pcls"] = self._num_pcls
        self._grp.attrs["mask_keys"] = self.masks._flush()
        self._grp.attrs["custom_keys"] = self.custom._flush()
        self._grp.attrs["custom_meta_keys"] = self.custom_meta._flush()
        self._proc._evt_idx += 1

    def _set_num_pcls(self, data: AnyVector) -> None:
        shape = data.shape
        num_pcls = shape[0]

        if self._num_pcls is None:
            self._num_pcls = num_pcls
        elif num_pcls != self._num_pcls:
            raise ValueError(
                    "Datasets within same event must have the same "
                    "number of particles (rows, if 2D array). "
                    f"Previous datasets had {self._num_pcls}, "
                    f"attempted dataset would have {num_pcls}."
                    )
        else:
            return

    @property
    def edges(self) -> AnyVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @edges.setter
    def edges(self, data: AnyVector) -> None:
        self._mk_dset(
                name="edges",
                data=data,
                shape=data.shape,
                dtype=self._types.edge,
                )
        self._num_edges = len(data)

    @property
    def edge_weights(self) -> DoubleVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @edge_weights.setter
    def edge_weights(self, data: DoubleVector) -> None:
        num_edges = self._num_edges
        num_weights = len(data)
        if (num_edges == 0) or (self._num_edges != num_weights):
            raise ValueError(
                f"Incompatible shapes. Number of edges = {num_edges} and "
                + f"number of weights = {num_weights}. Must be same size."
                )
        self._mk_dset(
                name="edge_weights",
                data=data,
                shape=data.shape,
                dtype=self._types.double,
                )

    @deprecated
    def set_edges(
            self,
            data: AnyVector,
            weights: Optional[DoubleVector] = None,
            strict_size: bool = True) -> None:
        """Write edge indices for event.

        :group: hepwrite
        
        Parameters
        ----------
        data : 2d array of ints
            Each row contains a pair of in / out edge indices.
        strict_size : bool
            If True will assume number of edges = number of particles
            (default: True).
        """
        if strict_size is True:
            self._set_num_pcls(data)
        if weights is not None:
            num_edges = len(data)
            num_weights = len(weights)
            if num_edges != num_weights:
                raise ValueError(
                    f"Incompatible shapes. Number of edges = {num_edges} and "
                    + f"number of weights = {num_weights}. Must be same size."
                    )
            self._mk_dset(
                    name='edge_weights',
                    data=weights,
                    shape=weights.shape,
                    dtype=self._types.double,
                    )
        self._mk_dset(
                name='edges',
                data=data,
                shape=data.shape,
                dtype=self._types.edge,
                )

    @deprecated
    def set_pmu(self, data: AnyVector) -> None:
        """Write 4-momentum for all particles to event.

        :group: hepwrite

        Parameters
        ----------
        data : 2d array of floats.
            Each row contains momentum of one particle,
            in order [px, py, pz, e].
        """
        self.pmu = data

    @property
    def pmu(self) -> AnyVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @pmu.setter
    def pmu(self, data: AnyVector) -> None:
        self._set_num_pcls(data)
        self._mk_dset(
                name='pmu',
                data=data,
                shape=data.shape,
                dtype=self._types.pmu,
                )

    @deprecated
    def set_color(self, data: AnyVector) -> None:
        """Write color / anticolor pairs for all particles to event.

        :group: hepwrite

        Parameters
        ----------
        data : 2d array of ints.
            Each row contains color / anticolor values respectively.
        """
        self.color = data

    @property
    def color(self) -> AnyVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @color.setter
    def color(self, data: AnyVector) -> None:
        self._set_num_pcls(data)
        self._mk_dset(
                name='color',
                data=data,
                shape=data.shape,
                dtype=self._types.color,
                )

    @deprecated
    def set_pdg(self, data: IntVector) -> None:
        """Pdg codes for all particles in event.

        :group: hepwrite

        Parameters
        ----------
        data : iterable or 1d numpy array
            Iterable of ints containing pdg codes for every
            particle in the event.
        """
        self.pdg = data

    @property
    def pdg(self) -> IntVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @pdg.setter
    def pdg(self, data: IntVector) -> None:
        self._set_num_pcls(data)
        self._mk_dset(
                name='pdg',
                data=data,
                shape=(self._num_pcls,),
                dtype=self._types.int,
                )

    @deprecated
    def set_status(self, data: IntVector) -> None:
        """Write status codes for all particles to event.

        :group: hepwrite
        
        Parameters
        ----------
        data : iterable or 1d numpy array
            Iterable of ints containing status codes for every
            particle in the event.
        """
        self.status = data

    @property
    def status(self) -> IntVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @status.setter
    def status(self, data: IntVector) -> None:
        self._set_num_pcls(data)
        self._mk_dset(
                name='status',
                data=data,
                shape=(self._num_pcls,),
                dtype=self._types.int,
                )

    @deprecated
    def set_helicity(self, data: HalfIntVector) -> None:
        """Write helicity values for all particles to event.

        :group: hepwrite

        Parameters
        ----------
        data : iterable or 1d numpy array
            Iterable of floats containing helicity values for every
            particle in the event.
        """
        self.helicity = data

    @property
    def helicity(self) -> HalfIntVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @helicity.setter
    def helicity(self, data: HalfIntVector) -> None:
        self._set_num_pcls(data)
        self._mk_dset(
                name='helicity',
                data=data,
                shape=(self._num_pcls,),
                dtype=self._types.helicity,
                )
    
    @deprecated
    def set_mask(self, name: str, data: BoolVector) -> None:
        """Write bool mask for all particles in event.

        :group: hepwrite
        
        Parameters
        ----------
        data : iterable or 1d numpy array
            Iterable of bools containing True / False values for
            every particle in event.
            Note: would also accept int iterable of 0s / 1s.

        Notes
        -----
        Example use cases:
            - identify particles from specific parent
            - provide mask for rapidity and pT cuts
            - if storing whole shower, identify final state
        """
        self.masks[name] = data

    @deprecated
    def set_custom_dataset(
            self, name: str,
            data: AnyVector,
            dtype: npt.DTypeLike,
            strict_size: bool = True) -> None:
        """Write a custom dataset to the event.

        :group: hepwrite

        Parameters
        ----------
        name : str
            Handle used when reading the data.
        data : nd numpy array
            data to store.
        dtype : valid string, numpy, or python data type
            Type in which your data should be encoded for
            storage.
            Note: using little Endian for builtin datasets.
        """
        if strict_size is True:
            self._set_num_pcls(data)
        self.custom[name] = data

    @deprecated
    def set_custom_meta(self, name: str, metadata: Any) -> None:
        """Store custom metadata to the event.

        :group: hepwrite

        Parameters
        ----------
        name : str
            Handle to access the metadata at read time.
        metadata : str, int, float, or iterables thereof
            The data you wish to store.
        """
        self.custom_meta[name] = metadata


@_export
class HdfProcessWriter(ProcessWriterBase):
    """Context manager interface to create and write processes.

    :group: hepwrite

    Attributes (write-only)
    -----------------------
    process_string : str
        String representing the hard process (eg. MadGraph format).
    signal_pdgs : numpy.ndarray[int]
        PDG codes within the hard process considered as "signal".
    com_energy : tuple[float, str]
        A tuple whose first element is the centre-of-mass collision
        energy for the hard process, and the second element is the unit
        that this energy is given in.
    custom_meta : MapWriter
        Write-only dictionary-like interface to set custom metadata to
        the process.

    Examples
    --------
    Wrap iterator with event_iter to obtain HdfEventWriter object, which
    stores pdg and pmu data to each event.

        >>> import heparchy
        >>> from showerpipe.generators import PythiaGenerator
        >>>
        >>> gen = PythiaGenerator("pythia-card.cmnd", "higgs.lhe.gz")
        >>> with heparchy.write.HdfWriter("showers.hdf5") as hep_file:
        >>>     with hep_file.new_process("higgs") as proc:
        >>>         for event_out, event_in in proc.event_iter(gen):
        >>>             # event_in is generated from Pythia
        >>>             event_out.pdg = event_in.pdg
        >>>             event_out.pmu = event_in.pmu

    """
    def __init__(self, file_obj: HdfWriter, key: str) -> None:
        self._file_obj = file_obj
        self.key = key
        self._evt_idx = 0
        self._grp: Group
        self.custom_meta: MapWriter[Any]

    def __enter__(self: HdfProcessWriter) -> HdfProcessWriter:
        self._grp = self._file_obj._buffer.create_group(self.key)
        self.custom_meta = MapWriter(self, _meta_setter)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        # count all of the events and write to attribute
        self._grp.attrs["custom_meta_keys"] = self.custom_meta._flush()
        self._grp.attrs["num_evts"] = self._evt_idx

    @deprecated
    def set_string(self, proc_str: str) -> None:
        """Writes the string formatted underlying event to the
        process metadata.

        :group: hepwrite
        
        Parameters
        ----------
        proc_str : str
            MadGraph formatted string representing the hard event,
            eg. p p > t t~
        """
        self.process_string = proc_str

    @property
    def process_string(self) -> str:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @process_string.setter
    def process_string(self, value: str) -> None:
        self._grp.attrs["process"] = value

    @deprecated
    def set_decay(self, in_pcls: Sequence[int], out_pcls: Sequence[int]
                  ) -> None:
        """Writes the pdgids of incoming and outgoing particles to
        process metadata.

        :group: hepwrite

        Parameters
        ----------
        in_pcls : tuple of ints
            pdgids of incoming particles.
            eg. for p p => (2212, 2212)
        out_pcls : tuple of ints
            pdgids of outgoing particles.
            eg. for t t~ => (6, -6)
        """
        self._grp.attrs['in_pcls'] = in_pcls
        self._grp.attrs['out_pcls'] = out_pcls

    @property
    def signal_pdgs(self) -> IntVector:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @signal_pdgs.setter
    def signal_pdgs(self, value: Sequence[int]) -> None:
        self._grp.attrs["signal_pdgs"] = value

    @deprecated
    def set_com_energy(self, energy: float, unit: str) -> None:
        """Writes the CoM energy and unit for the collision to
        process metadata.

        :group: hepwrite

        Parameters
        ----------
        energy : float
            eg. 13.0
        out_pcls : str
            eg. 'GeV'
        """
        self._grp.attrs['com_e'] = energy
        self._grp.attrs['e_unit'] = unit

    @property
    def com_energy(self) -> Tuple[float, str]:
        raise WriteOnlyError(_WRITE_ONLY_MSG)

    @com_energy.setter
    def com_energy(self, value: Tuple[float, str]) -> None:
        self._grp.attrs["com_e"] = value[0]
        self._grp.attrs["e_unit"] = value[1]

    @deprecated
    def set_custom_meta(self, name: str, metadata: Any) -> None:
        """Store custom metadata to the process.

        :group: hepwrite

        Parameters
        ----------
        name : str
            Handle to access the metadata at read time.
        metadata : str, int, float, or iterables thereof
            The data you wish to store.
        """
        self._grp.attrs[name] = metadata

    def new_event(self) -> HdfEventWriter:
        return HdfEventWriter(self)

    def event_iter(self, iterable: Iterable[IterItem]
                   ) -> Iterator[Tuple[HdfEventWriter, IterItem]]:
        """Wraps an iteratable object, returning a new iterator which
        yields a new writeable event object followed by the value
        obtained from the passed iterator.

        :group: hepwrite

        Parameters
        ----------
        iterable : Iterable[IterItem]
            Any iterable object, eg. tuples, lists, generators, etc.

        Returns
        -------
        event : HdfEventWriter
            Writeable event object.
        value : IterItem
            The value yielded by the input iterable.
        """
        for value in iterable:
            with self.new_event() as event:
                yield event, value


@_export
class HdfWriter(WriterBase):
    """Create a new heparchy hdf5 file object with write access.

    :group: hepwrite

    Parameters
    ----------
    path : Path | str
        Filepath for output.
    compression : Compression | str
        Supports "gzip" or "lzf" compression for datasets.
    compression_level : int, optional
        Integer between 1 - 9 setting gzip compression factor. Ignored
        if compression type is lzf.

    Examples
    --------
    Open a new heparchy HDF5 file, and create a process within it:

        >>> import heparchy
        >>>
        >>> with heparchy.write.HdfWriter("showers.hdf5") as hep_file:
        >>>     with hep_file.new_process("higgs") as proc:
        >>>         ...
    """
    def __init__(
            self,
            path: Union[Path, str],
            compression: Union[str, Compression] = Compression.GZIP,
            compression_level: Optional[int] = 4,
            ) -> None:
        self.path = Path(path)
        self._buffer: File
        if isinstance(compression, str):
            compression = Compression(compression.lower())
        self._cmprs = compression
        if compression is Compression.LZF:
            compression_level = None
        self._cmprs_lvl = compression_level

    def __enter__(self: HdfWriter) -> HdfWriter:
        self._buffer = h5py.File(self.path, 'w', libver='latest')
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self._buffer.attrs["layout"] = "heparchy"
        self._buffer.attrs["version_tuple"] = tuple(
                map(str, hrc.__version_tuple__))
        self._buffer.attrs["version"] = hrc.__version__
        self._buffer.close()

    def new_process(self, name: str) -> HdfProcessWriter:
        """Returns a context handler object for storage of data in
        a given hard process.

        Events can be iteratively added to this process by repeatedly
        calling the ``new_event()`` method, which itself returns a
        context handler object.

        :group: hepwrite

        Parameters
        ----------
        key : str
            Arbitrary name, used to look up data with reader.

        Returns
        -------
        process : HdfProcessWriter
            Context manager for creating processes within the file,
            providing methods for adding metadata and events.
            See heparchy.write.HdfProcessWriter for more.
        """
        return HdfProcessWriter(self, key=name)



"""
`heparchy.read`
===============

Interfaces to read from standard particle physics formats,
and the heparchy formats.

Notes
-----
Currently only implemented for hepmc and hdf5.
"""
from . import hdf
from .hdf import *
import heparchy.read.hepmc as hepmc


HepMC = hepmc.HepMC


class Wiggle:
    pass


__all__ = ["HepMC", "Wiggle"]
__all__.extend(hdf.__all__)



"""
heparchy.data.event
===================

This module defines the dataclasses and methods which construct the
DAG of the event generation.

Deprecation notice
------------------
Warning: this functionality is going to be moved from the heparchy
package into graphicle, instead.

Graphicle is a package designed to embed and process particle physics
data on graph stuctures, whereas heparchy is intended only to provide
hierarchical and high performance I/O for particle physics data.
"""

from copy import deepcopy
from functools import wraps

import attr
import numpy as np

import heparchy.utils as utils


def val_elems_int(instance, attribute, value):
    all_ints = set(map(type, value)) == {int} # True if elems all ints
    empty = False
    if not value:
        empty = True
    if not empty and not all_ints:
        raise TypeError(
                f'Inputs must be iterables of integers. Got value={value}')

def val_int_array(instance, attribute, value):
    pass

@attr.s(kw_only=True, frozen=True, slots=True)
class SignalVertex:
    """Data structure to define which of the vertices represent a
    signal in an event, and which of the given vertex's descendants
    are to be followed in the subsequent showering.

    Each of the inputs are to be formatted as an iterable of integers,
    representing the pdgs of the particles.

    Keyword arguments:
        incoming: particles incident on the vertex
        outgoing: particles outbound from the vertex
        follow: outbound particles marked to have their children tracked
    """

    from typing import Set as __Set

    __PdgSet = __Set[int]
    __pdg_kwargs = dict(converter=set, validator=[val_elems_int])
    incoming: __PdgSet = attr.ib(**__pdg_kwargs)
    outgoing: __PdgSet = attr.ib(**__pdg_kwargs)
    follow: __PdgSet = attr.ib(**__pdg_kwargs)

# @attr.s
# class ParticleSet:
#     pmu: np.ndarray = attr.ib(
#             **__array_kwargs)
#     pdg: np.ndarray = attr.ib(
#             **__array_kwargs)
#     __pmu_vector = self.__vector_view()

#     def __vector_view(self):
#         dtype = deepcopy(self.pmu.dtype)
#         dtype.names = ('x', 'y', 'z', 't')
#         pmu_vec = self.pmu.view(dtype).view(self.__vector.MomentumNumpy4D)
#         return pmu_vec

#     @property
#     def pt(self):
#         return self.__pmu_vector.pt

#     @property
#     def eta(self):
#         return self.__pmu_vector.eta

@attr.s(on_setattr=attr.setters.convert)
class ShowerData:
    import pandas as __pd
    import networkx as __nx
    import vector as __vector
    from typicle import Types as __Types

    __types = __Types()
    __array_kwargs = dict(
            eq=attr.cmp_using(eq=np.array_equal),
            )
    edges: np.ndarray = attr.ib(
            converter=utils.structure_edges,
            **__array_kwargs)
    pmu: np.ndarray = attr.ib(
            converter=utils.structure_pmu,
            **__array_kwargs)
    pdg: np.ndarray = attr.ib(
            **__array_kwargs)
    final: np.ndarray = attr.ib(
            **__array_kwargs)

    @classmethod
    def empty(cls):
        return cls(
            edges = np.array([[0, 1]], dtype=cls.__types.int),
            pmu=np.array([[0.0, 0.0, 0.0, 0.0]], dtype=cls.__types.pmu[0][1]),
            pdg=np.array([1], dtype=cls.__types.int),
            final=np.array([False], dtype=cls.__types.bool)
            )

    @property
    def __pmu_vector(self):
        dtype = deepcopy(self.pmu.dtype)
        dtype.names = ('x', 'y', 'z', 't')
        pmu_vec = self.pmu.view(dtype).view(self.__vector.MomentumNumpy4D)
        return pmu_vec

    def pt(self, mask=None):
        pmu_subset = self.__pmu_vector[mask].squeeze()
        return pmu_subset.pt

    def eta(self, mask=None):
        pmu_subset = self.__pmu_vector[mask].squeeze()
        return pmu_subset.eta

    def phi(self, mask=None):
        pmu_subset = self.__pmu_vector[mask].squeeze()
        return pmu_subset.phi

    def flush_cache(self):
        """Clears cached private attributes of ShowerData instance.
        These must currently be cleared manually after updating
        public attributes of an instance.

        Notes
        -----
        If looping through many events, performance is vastly improved
        if only one instance of the ShowerData class is created,
        and the attributes of that instance are iteratively updated.
        In this case, it is necessary to use `flush_cache` during each
        iteration, so the next event does not have cached data from the
        previous one.

        In a future release, this will be handled automatically whenever
        an attribute value is updated.
        """
        try:
            self.__shower
        except AttributeError:
            pass
        else:
            del self.__shower

    def to_networkx(self, data=['pdg']):
        """Output directed acyclic graph representation of the shower,
        implemented by NetworkX. Each edge is a particle, and each node
        is an interaction vertex, except for the terminating leaf nodes.

        Parameters
        ----------
        data : iterable of strings
            Specify which of the particle properties should be embedded
            on the edges.
            Valid options are:
                - pdg [default]
                - pmu
                - final
        """
        # form edges with pdg data on for easier ancestry tracking
        names = data
        data_rows = (getattr(self, name) for name in names)
        data_rows = zip(*data_rows)
        edge_dicts = (dict(zip(names, row)) for row in data_rows)
        edges = zip(self.edges['in'], self.edges['out'], edge_dicts)
        shower = self.__nx.DiGraph()
        shower.add_edges_from(edges)
        return shower

    def __requires_shower(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            try:
                self.__shower
            except AttributeError:
                self.__shower = self.to_networkx()
            return func(self, *args, **kwargs)
        return inner

    def to_pandas(self, data=('pdg', 'final')):
        """Return event data in a single pandas dataframe.

        Parameters
        ----------
        data : iterable of strings or True
            The particle properties to include as columns.
            Valid options are:
                - pdg [default]
                - final [default]
                - pmu
                - x, y, z, or e
                - edges
                - pt
                - eta
                - phi
            If set to True instead of iterable, all data are included.

        Notes
        -----
        This is particularly useful if you want to mask the data
        in complex ways involving multiple particle properties at once,
        using the dataframe's `query` method.
        """
        # collect column data into a dict, ready to pass to pandas
        cols = {
            'pdg': self.pdg,
            'final': self.final,
            'edge_in': self.edges['in'],
            'edge_out': self.edges['out'],
            'pt': self.pt(),
            'eta': self.eta(),
            'phi': self.phi(),
            }
        pmu_cols = list(self.pmu.dtype.names)
        cols.update({col_name: self.pmu[col_name] for col_name in pmu_cols})
        # check which data to include
        if data == True: # all of it
            return self.__pd.DataFrame(cols)
        else: # restricted selection
            try:
                iterator = iter(data)
            except TypeError:
                print("data must either be an iterable, or True")
            else:
                # define valid input
                col_names_edge = ['edge_in', 'edge_out']
                valid_col_names = list(cols.keys())
                valid_col_names = set(valid_col_names + col_names_edge
                                      + ['pmu', 'edges'])
                # validate user input
                user_col_names = set(data)
                invalid_col_names = user_col_names.difference(valid_col_names)
                # discard invalid input
                col_names = user_col_names.intersection(valid_col_names)
                col_names = list(col_names)
                if invalid_col_names: # let user know of invalid input
                    print(f'Warning: {len(invalid_col_names)} invalid column '
                          + f'name(s) provided, {invalid_col_names}. '
                          + 'Omitting.')
                if 'edges' in col_names: # convert abbreviations
                    col_names.remove('edges')
                    col_names += col_names_edge
                if 'pmu' in col_names:
                    col_names.remove('pmu')
                    col_names += pmu_cols
                return self.__pd.DataFrame( # populate dataframe with selection
                        {key: cols[key] for key in col_names})

    @__requires_shower
    def vertex_pdg(self):
        """Returns a generator object which loops over all interaction
        vertices in the event, yielding the vertex id, and pdg codes
        of incoming and outgoing particles to the vertex, respectively.
        """
        shower = self.__shower
        for vertex in shower:
            incoming = shower.in_edges(vertex, data='pdg')
            outgoing = shower.out_edges(vertex, data='pdg')
            if incoming and outgoing:
                vtxs_in, _, pdgs_in = zip(*incoming)
                _, vtxs_out, pdgs_out = zip(*outgoing)
                yield vertex, set(pdgs_in), set(pdgs_out)
    
    @__requires_shower
    def signal_mask(self, signal_vertices):
        """Locates given vertices, based on their incoming and outgoing
        particles.

        Parameters
        ----------
        signal_vertices : iterable of `heparchy.SignalVertex` objects
            Specify multiple vertices to locate, and which of their
            descendants to follow.
            See `heparchy.SignalVertex` for information.

        Returns
        -------
        signal_masks : list of dicts containing 1d boolean numpy arrays
            Masks over all particles in the event, identifying which
            descend from the followed particle, produced at the
            chosen vertex.
            The output for each vertex is returned, in the same order as
            provided, as an element in a list.
            These elements are dictionaries, whose keys refer to the pdg
            codes of the followed particles, and whose values are the
            boolean arrays which identify their descendants.
        """
        # --- LOCATE SIGNAL VERTICES IN SHOWER --- #
        is_signal = []
        # mark where the signal vertices are in the shower:
        for vertex, pdgs_in, pdgs_out in self.vertex_pdg():
            is_signal.append(tuple(
                [vertex] + [sig_vtx.incoming.issubset(pdgs_in)
                            and sig_vtx.outgoing.issubset(pdgs_out)
                            for sig_vtx in signal_vertices]
                ))
        # split into two iterables of vertex ids and booleans, respectively
        is_signal = list(zip(*is_signal))
        vertices = is_signal[0]
        signal_lists = is_signal[1:]
        # --- POPULATE A LIST OF DICTIONARIES WITH THE SIGNAL MASKS --- #
        signal_masks = []
        # iterate through id and corresponding SignalVertex object together
        for signal_list, signal_vertex in zip(signal_lists, signal_vertices):
            follow_masks = dict() # to hold the mask, keyed by pdg
            if True in signal_list:
                # find the vertex id of the signal vertices:
                signal_id = vertices[signal_list.index(True)]
                # identify which vertex products need to be followed:
                edges_out_info = self.__shower.out_edges(signal_id, data='pdg')
                follow_edges = filter(
                        lambda edge: edge[-1] in signal_vertex.follow,
                        edges_out_info
                        )
                # store incident vtx of followed pcl as follow_id
                _, follow_ids, follow_pdgs = zip(*follow_edges)
                # find the descendants of follow_id vtxs for each followed pcl
                for follow_id, follow_pdg in zip(follow_ids, follow_pdgs):
                    desc_vtxs = self.__nx.descendants(self.__shower, follow_id)
                    # convert into structured np array for set comparison
                    desc_edges = list(self.__shower.edges(nbunch=desc_vtxs))
                    desc_edges = utils.structure_edges(np.array(desc_edges))
                    # create mask identifying edges / pcls desc from signal
                    mask = np.isin(self.edges, desc_edges, assume_unique=False)
                    follow_masks.update({follow_pdg: mask}) # store to dict
            else: # if vertex not located, fill with all False masks
                mask = np.full_like(self.final, fill_value=False)
                follow_masks.update(
                    {follow_pdg: mask for follow_pdg in signal_vertex.follow})
            signal_masks += [follow_masks]
        return signal_masks

    def copy(self):
        """Returns a copy of the ShowerData instance as a new object
        in memory.
        """
        return deepcopy(self)




from abc import ABC, abstractmethod
from typing import Any

import numpy as np
from h5py import Group, File


class EventWriterBase(ABC):
    @abstractmethod
    def set_edges(self, data: np.ndarray) -> None:
        pass

    @abstractmethod
    def set_pmu(self, data: np.ndarray) -> None:
        pass

    @abstractmethod
    def set_color(self, data: np.ndarray) -> None:
        pass

    @abstractmethod
    def set_pdg(self, data: np.ndarray) -> None:
        pass

    @abstractmethod
    def set_mask(self, name: str, data: np.ndarray) -> None:
        pass

    @abstractmethod
    def set_custom_dataset(
            self, name: str, data: np.ndarray, dtype: Any) -> None:
        pass

class ProcessWriterBase(ABC):
    @abstractmethod
    def set_string(self, proc_str: str) -> None:
        pass

    @abstractmethod
    def set_decay(self, in_pcls: tuple, out_pcls: tuple) -> None:
        pass

    @abstractmethod
    def set_com_energy(self, energy: float, unit: str) -> None:
        pass

    @abstractmethod
    def set_custom_meta(self, name: str, metadata: Any) -> None:
        pass

    @abstractmethod
    def new_event(self) -> EventWriterBase:
        pass


class WriterBase(ABC):
    @abstractmethod
    def new_process(self, name: str) -> ProcessWriterBase:
        pass



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List, Any, Dict

import numpy as np
import numpy.typing as npt


DoubleVector = npt.NDArray[np.float64]
IntVector = npt.NDArray[np.int32]
HalfIntVector = npt.NDArray[np.int16]
BoolVector = npt.NDArray[np.bool_]


class EventReaderBase(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        """Getter for event name"""

    @property
    @abstractmethod
    def count(self) -> int:
        """Getter for number of particles in event"""

    @property
    @abstractmethod
    def edges(self) -> np.ndarray:
        """Getter for edges in event generation DAG."""

    @property
    @abstractmethod
    def pmu(self) -> np.ndarray:
        """Getter for 4-momentum of all particles in event."""

    @property
    @abstractmethod
    def color(self) -> np.ndarray:
        """Getter for color pairs of all particles in event."""

    @property
    @abstractmethod
    def pdg(self) -> IntVector:
        """Getter for pdg codes of all particles in event."""

    @property
    @abstractmethod
    def final(self) -> BoolVector:
        """Getter for mask identifying final state particles."""

    @property
    @abstractmethod
    def available(self) -> List[str]:
        """Provides list of all dataset names in event."""

    @abstractmethod
    def mask(self, name: str) -> BoolVector:
        """Getter for a mask of a given name over all particles in event."""

    @abstractmethod
    def get_custom(self, name: str) -> Any:
        """Getter for user-defined dataset stored in event."""

    @abstractmethod
    def copy(self) -> EventReaderBase:
        """Returns a deepcopy of this dataclass instance."""


class ProcessReaderBase(ABC):
    """An iterator of EventReaderBase dataclass instances over the
    events nested in the process.
    """

    @property
    @abstractmethod
    def string(self) -> str:
        """Getter for the MadGraph style process string."""

    @property
    @abstractmethod
    def decay(self) -> Dict[str, IntVector]:
        """Returns dictionary with two entries, describing the hard
        interaction for this process.

        Dictionary items
        ----------------
        in_pcls : np.ndarray
            The pdg codes of the incoming particles.
        out_pcls : np.ndarray
            The pdg codes of the outgoing particles.
        """

    @property
    @abstractmethod
    def com_energy(self) -> ComEnergyType:
        """Returns dictionary with two entries, describing the
        centre-of-mass energy for this hard process.

        Dictionary items
        ----------------
        energy : float
            The value of the centre-of-mass energy.
        unit : string
            The unit of energy, eg. GeV.
        """

    @abstractmethod
    def get_custom_meta(self, name: str) -> Any:
        """Returns user-defined piece of metadata."""

    @abstractmethod
    def read_event(self, evt_num: int) -> EventReaderBase:
        """Alias for subscripting to prevent breaking changes.
        Remove at 1.0 release.
        """

    @abstractmethod
    def __getitem__(self, evt_num: int) -> EventReaderBase:
        """Provides option to get EventReaderBase object without
        iteration. Instead user supplies event number.
        """


class ReaderBase(ABC):
    """Context manager for opening heparchy formatted files."""

    @abstractmethod
    def read_process(self, name: str) -> ProcessReaderBase:
        """Returns process reader iterator object."""



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath("."))
from importlib.metadata import version as get_version
from typing import Optional, Dict, Any
from pathlib import Path
import inspect

import heparchy as proj_pkg
from pygit2 import Repository


# -- Project information -----------------------------------------------------

project = "heparchy"
copyright = "2022, Jacan Chaplais"
author = "Jacan Chaplais"
url = "https://github.com/jacanchaplais/heparchy"
repo = Repository(".")
commit_hash = repo.head.target
proj_dir = Path(repo.path).parent

# The full version, including alpha/beta/rc tags
release = get_version(project)
# for example take major/minor
version = ".".join(release.split(".")[:2])

packages = set(map(Path, ("heparchy", "heparchy/read", "heparchy/write",
                          "heparchy/data")))


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_mdinclude",
    "sphinx_immaterial",
    "sphinx_immaterial.apidoc.python.apigen",
    "sphinx.ext.linkcode"
]

autoapi_dirs = [proj_dir]
autodoc_default_options = {
    "imported-members": True,
    "members": True,
    "special-members": True,
    # "inherited-members": "ndarray",
    # "member-order": "groupwise",
}
autodoc_typehints = "signature"
autodoc_typehints_description_target = "documented"
autodoc_typehints_format = "short"
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]


def get_symbol(root_symbol: Any, name: str) -> Any:
    name_chunks = name.split(".", 1)
    if len(name_chunks) > 1:
        parent_name, child_name = name_chunks
        parent_symbol = get_symbol(
                getattr(root_symbol, parent_name), child_name)
    else:
        parent_name = name_chunks[0]
        parent_symbol = getattr(root_symbol, parent_name)
    return parent_symbol


def get_module_name(root_symbol: Any, name: str) -> str:
    symbol = get_symbol(root_symbol, name)
    try:
        module_name = symbol.__module__
    except AttributeError:
        parent_name = ".".join(name.split(".")[:-1])
        module_name = get_module_name(root_symbol, parent_name)
    return module_name


def linkcode_resolve(domain: str, info: Dict[str, str]) -> Optional[str]:
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = Path(info["module"].replace(".", "/") + ".py")
    if not (proj_dir / filename).exists():
        symbol_name = (info["module"].removeprefix(f"{project}.")
                       + "." + info["fullname"])
        try:
            module_name = get_module_name(proj_pkg, symbol_name)
            filename = Path(module_name.replace(".", "/") + ".py")
        except AttributeError:
            return None
        if filename in packages:
            filename = filename / "__init__.py"
    link = f"{url}/blob/{commit_hash}/{filename}"
    return link


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".ipynb_checkpoints"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = "alabaster"
html_theme = "sphinx_immaterial"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for EPUB output
epub_show_urls = "footnote"

html_theme_options = {
    "icon": {
        "repo": "fontawesome/brands/github",
    },
    "site_url": f"https://{project}.readthedocs.io/",
    "repo_url": url,
    "repo_name": f"jacanchaplais/{project}",
    "repo_type": "github",
    "social": [
        {
            "icon": "fontawesome/brands/github",
            "link": url,
        },
        {
            "icon": "fontawesome/brands/python",
            "link": f"https://pypi.org/project/{project}/",
        },
    ],
    "edit_uri": "",
    "globaltoc_collapse": False,
    "features": [
        # "navigation.expand",
        "navigation.tabs",
        # "toc.integrate",
        # "navigation.sections",
        # "navigation.instant",
        # "header.autohide",
        "navigation.top",
        "navigation.tracking",
        "toc.follow",
        "toc.sticky",
    ],
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "accent": "deep-orange",
            "toggle": {
                "icon": "material/weather-night",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "accent": "deep-orange",
            "toggle": {
                "icon": "material/weather-sunny",
                "name": "Switch to light mode",
            },
        },
    ],
    "analytics": {"provider": "google", "property": "G-4FW9NCNFZH"},
    "version_dropdown": True,
    "version_json": "../versions.json",
}

html_last_updated_fmt = ""
html_use_index = True
html_domain_indices = True

epub_show_urls = "footnote"


# Python apigen configuration
python_apigen_modules = {"heparchy": "api/heparchy."}

modules = inspect.getmembers(proj_pkg, inspect.ismodule)
for module in modules:
    python_apigen_modules[f"{project}." + module[0]] = (
        f"api/{project}." + module[0] + "."
    )



"""
`heparchy.read.hdf`
===================

Defines the data structures to open and iterate through heparchy
formatted HDF5 files.
"""
from __future__ import annotations
from pathlib import Path
from os.path import basename
from copy import deepcopy
from functools import cached_property
from collections.abc import Mapping
from typing import (
        Any, List, Dict, Sequence, Type, Iterator, Union, Set,
        TypeVar, Callable, Generic, Tuple, Optional)

import numpy as np
import h5py
from h5py import AttributeManager, Group, Dataset

from heparchy.utils import event_key_format, deprecated
from .base import ReaderBase, EventReaderBase, ProcessReaderBase
from heparchy.annotate import (
        IntVector, HalfIntVector, DoubleVector, BoolVector, AnyVector,
        DataType)


__all__: List[str] = []


def _export(procedure: ExportType) -> ExportType:
    __all__.append(procedure.__name__)
    return procedure


_NOT_NUMPY_ERR = ValueError("Stored data type is corrupted.")
_NO_EVENT_ERR = AttributeError("Event reader not pointing to event.")
_BUILTIN_PROPS: Set[str] = set()
_BUILTIN_METADATA = {  # TODO: work out non-hardcoded
    "HdfEventReader": {"num_pcls", "mask_keys"},
    "HdfProcessReader": {"signal_pdgs", "process", "com_e", "e_unit"}
    }

MetaDictType = Dict[str, Union[str, int, float, bool, AnyVector]]
ReaderType = Union["HdfEventReader", "HdfProcessReader"]
ExportType = TypeVar("ExportType", Callable, Type)
PropMethod = TypeVar("PropMethod", bound=Callable)
MapValue = TypeVar("MapValue")


@_export
class ReadOnlyError(RuntimeError):
    """Raised when trying to write to read-only data.
    :group: hepread
    """


def _reg_event_builtin(func: PropMethod) -> PropMethod:
    _BUILTIN_PROPS.add(func.__name__)
    return func


def _stored_keys(attrs: AttributeManager, key_attr_name: str) -> Iterator[str]:
    key_ds = attrs[key_attr_name]
    if not isinstance(key_ds, np.ndarray):
        raise _NOT_NUMPY_ERR
    for name in tuple(key_ds):
        yield name


def _grp_key_iter(reader: Group) -> Iterator[str]:
    yield from reader.keys()


def _meta_iter(reader: Group) -> Iterator[str]:
    key_attr_name = "custom_meta_keys"
    if key_attr_name not in reader.attrs:
        names = set(reader.attrs.keys())
        for name in (names - _BUILTIN_METADATA[reader.__class__.__name__]):
            yield name
    else:
        yield from _stored_keys(reader.attrs, key_attr_name)


@_export
class MapReader(Generic[MapValue], Mapping[str, MapValue]):
    """Read-only dictionary-like interface to user-named heparchy
    datasets.

    :group: hepread

    Parameters
    ----------
    reader : HdfProcessReader | HdfEventReader
        The reader instance to which the MapReader will be attached.
    iter_func : callable
        The strategy used to iterate through the keys for the MapReader
        interface.

    Raises
    ------
    ReadOnlyError
        Dictionary-like behaviour which requires write-access for
        datasets, as access to the data is strictly read-only.

    Notes
    -----
    This is a concrete subclass of collections.abc.MutableMapping, and
    thus includes all functionality expected of a dictionary, including
    iteration over keys and values, subscriptable access to datasets,
    etc.
    """
    def __init__(self,
                 reader: ReaderType,
                 grp_attr_name: str,
                 iter_func: Callable[..., Iterator[str]]) -> None:
        self._reader = reader
        self._grp_attr_name = grp_attr_name
        self._iter_func = iter_func

    @property
    def _grp(self) -> Group:
        return getattr(self._reader, self._grp_attr_name)

    def __repr__(self) -> str:
        dset_repr = "<Read-Only Data>"
        kv = ", ".join(map(lambda k: f"\'{k}\': {dset_repr}", self))
        return "{" + f"{kv}" + "}"

    def __len__(self) -> int:
        return len(tuple(iter(self)))

    def __getitem__(self, name: str) -> MapValue:
        if name not in set(self):
            raise KeyError("No data stored with this name")
        if self._iter_func.__name__ == "_meta_iter":
            return self._grp.attrs[name]  # type: ignore
        data = self._grp[name]
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]  # type: ignore

    def __setitem__(self, name: str, data: MapValue) -> None:
        raise ReadOnlyError("Value is read-only")

    def __delitem__(self, name: str) -> None:
        raise ReadOnlyError("Value is read-only")

    def __iter__(self) -> Iterator[str]:
        yield from self._iter_func(self._grp)


def _type_error_str(data: Any, dtype: type) -> str:
    return ("Type mismatch: retrieved value should be of "
            f"type {dtype}, found instance of {type(data)}")


def _check_type(data: Any, dtype: Type[DataType]) -> DataType:
    if not isinstance(data, dtype):
        raise TypeError(_type_error_str(data, dtype))
    return data


def _format_array_dict(
        keys: Sequence[str], in_dict: Dict[str, Any]) -> Dict[str, AnyVector]:
    array_dict = dict()
    for key in keys:
        val = in_dict[key][...]
        if not isinstance(val, np.ndarray):
            raise TypeError(_type_error_str(val, np.ndarray))
        array_dict[key] = val
    return array_dict


@_export
class HdfEventReader(EventReaderBase):
    """A heparchy event reader object.

    :group: hepread

    Attributes
    ----------
    name : str
        The string representation of the event's name.
    count : int
        The number of particles within the event.
    edges : numpy.ndarray[int]
        A structured array providing edges for a graph representation,
        with fields "in" and "out".
    edge_weights : numpy.ndarray[double]
        Weights attributed to each edge.
    pmu : numpy.ndarray[double]
        A structured array providing particle 4-momenta, with fields
        "x", "y", "z", "e"
    color : numpy.ndarray[int]
        A structured array with fields "color" and "anti-color".
    pdg : numpy.ndarray[int]
        PDG codes providing the identity of every particle in the event.
    status : numpy.ndarray[int]
        Status codes, describing the reason for the generation of each
        particle in the event.
    helicity : numpy.ndarray[int]
        Spin polarisations of every particle in the event.
    final : numpy.ndarray[bool], deprecated
        Boolean mask over the data identifying particles in final state.
        Deprecated, simply use masks attribute with "final" subscript.
    available : list[str]
        Names of the datasets stored within the event.
    masks : MapReader[numpy.ndarray[bool]]
        Read-only dictionary-like interface to access boolean masks over
        the particle datasets.
    custom : MapReader[numpy.ndarray]
        Read-only dictionary-like interface to access user-defined
        custom datasets.
    custom_meta : MapReader
        Read-only dictionary-like interface to access user-defined
        metadata on the event.
    """
    def __init__(self) -> None:
        self.__name: Optional[str] = None
        self.__grp: Optional[Group] = None
        self._custom_grp: Group
        self._mask_grp: Group

    @property
    def _name(self) -> str:
        if self.__name is None:
            raise _NO_EVENT_ERR
        return self.__name

    @_name.setter
    def _name(self, val: str) -> None:
        self.__name = val

    @property
    def _grp(self) -> Group:
        if self.__grp is None:
            raise _NO_EVENT_ERR
        return self.__grp

    @_grp.setter
    def _grp(self, val: Group) -> None:
        self.__grp = val
        custom = val["custom"]
        mask = val["masks"]
        if (not isinstance(custom, Group)) or (not isinstance(mask, Group)):
            raise ValueError("Group not found")
        self._custom_grp = custom
        self._mask_grp = mask

    @property
    def name(self) -> str:
        return str(basename(self._name))

    @property
    def count(self) -> int:
        value = self._grp.attrs['num_pcls']
        if not isinstance(value, np.integer):
            raise _NOT_NUMPY_ERR
        return int(value)

    @property  # type: ignore
    @_reg_event_builtin
    def edges(self) -> AnyVector:
        data = self._grp["edges"]
        if not isinstance(data, Dataset):
            raise ValueError("Stored data is corrupted")
        return data[...]

    @property  # type: ignore
    @_reg_event_builtin
    def edge_weights(self) -> DoubleVector:
        data = self._grp['edge_weights']
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]

    @property  # type: ignore
    @_reg_event_builtin
    def pmu(self) -> AnyVector:
        data = self._grp['pmu']
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]

    @property  # type: ignore
    @_reg_event_builtin
    def color(self) -> AnyVector:
        data = self._grp['color']
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]

    @property  # type: ignore
    @_reg_event_builtin
    def pdg(self) -> IntVector:
        data = self._grp['pdg']
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]

    @property  # type: ignore
    @_reg_event_builtin
    def status(self) -> HalfIntVector:
        data = self._grp['status']
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]

    @property  # type: ignore
    @_reg_event_builtin
    def helicity(self) -> HalfIntVector:
        data = self._grp['helicity']
        if not isinstance(data, Dataset):
            raise _NOT_NUMPY_ERR
        return data[...]

    @property  # type: ignore
    @deprecated
    def final(self) -> BoolVector:
        return self.masks["final"]

    @property
    def available(self) -> List[str]:
        dataset_names = list()
        self._grp.visit(lambda name: dataset_names.append(name))
        return dataset_names

    @deprecated
    def mask(self, name: str) -> BoolVector:
        """Returns a boolean mask stored with the provided name."""
        return self.masks[name]

    @cached_property
    def masks(self) -> MapReader[BoolVector]:
        return MapReader[BoolVector](self, "_mask_grp", _grp_key_iter)

    @deprecated
    def get_custom(self, name: str) -> AnyVector:
        """Returns a custom dataset stored with the provided name."""
        return self.custom[name]
    
    @cached_property
    def custom(self) -> MapReader[AnyVector]:
        return MapReader[AnyVector](self, "_custom_grp", _grp_key_iter)

    @deprecated
    def get_custom_meta(self, name: str) -> Any:
        """Returns custom metadata stored with the provided name."""
        return self._grp.attrs[name]

    @cached_property
    def custom_meta(self) -> MapReader[Any]:
        return MapReader[Any](self, "_grp", _meta_iter)

    def copy(self) -> HdfEventReader:
        """Returns a deep copy of the event object."""
        return deepcopy(self)


@_export
class HdfProcessReader(ProcessReaderBase):
    """A heparchy process object. This is a numerically subscriptable
    iterator.

    :group: hepread

    Parameters
    ----------
    file_obj : HdfReader
        Instantiated heparchy file object.
    key : str
        The name of the process to be opened.

    Attributes
    ----------
    process_string : str
        The MadGraph string representation of the process.
    string : str, deprecated
        Deprecated, use process_string instead.
    signal_pdgs : numpy.ndarray[int]
        The PDG codes in the hard process which are considered signal.
    decay : dict[str, numpy.ndarray[int]], deprecated
        Provides access to incoming and outgoing parton PDG codes in
        the process. This has been deprecated due to the complexity
        introduced with intermediate states. Consider using signal_pdgs
        instead.
    com_energy : tuple[float, str]
        The magnitude and unit of the collision energy.
    custom_meta : MapReader
        Dictionary-like access to user-stored metadata.

    Examples
    --------
    Access a process named "higgs" and access events:

        >>> with heparchy.read.HdfReader("showers.hdf5") as hep_file:
        >>>     proc = hep_file["higgs"]
        >>>     for event in proc:  # iterate with for loop
        >>>         ...
        >>>     event = proc[5]  # 6th event with explicit indexing

    Read metadata stored in a process:

        >>> with heparchy.read.HdfReader("showers.hdf5") as hep_file:
        >>>     proc = hep_file["higgs"]
        >>>     print(proc.process_string)
        >>>     print(proc.signal_pdgs)
        >>>     print(proc.com_energy)
        >>>     print(proc.custom_meta["decay_channels"])
            p p > h z , (h > b b~) , (z > l+ l-)
            [25, 5, -5]
            (1300.0, "GeV")
            semileptonic
    
    """
    def __init__(self, file_obj: HdfReader, key: str) -> None:
        self._evt = HdfEventReader()
        grp = file_obj._buffer[key]
        if not isinstance(grp, Group):
            raise KeyError(f"{key} is not a process")
        self._grp: Group = grp
        self._meta: MetaDictType = dict(file_obj._buffer[key].attrs)
        self.custom_meta: MapReader[Any] = MapReader(self, "_grp", _meta_iter)

    def __len__(self) -> int:
        return int(self._meta["num_evts"])

    def __iter__(self) -> Iterator[HdfEventReader]:
        self._evt_gen = (self[i] for i in range(len(self)))
        return self

    def __next__(self) -> HdfEventReader:
        return next(self._evt_gen)

    def __getitem__(self, evt_num: int) -> HdfEventReader:
        evt_name = event_key_format(evt_num)
        self._evt._name = evt_name
        evt_grp = self._grp[evt_name]
        if not isinstance(evt_grp, Group):
            raise ValueError
        self._evt._grp = evt_grp
        return self._evt

    @deprecated
    def read_event(self, evt_num: int) -> HdfEventReader:
        """Returns the event indexed by the passed index."""
        return self[evt_num]

    @property  # type: ignore
    @deprecated
    def string(self) -> str:
        return self.process_string

    @property
    def process_string(self) -> str:
        return _check_type(self._meta["process"], str)

    @property  # type: ignore
    @deprecated
    def decay(self) -> Dict[str, IntVector]:
        return _format_array_dict(("in_pcls", "out_pcls"), self._meta)

    @property
    def signal_pdgs(self) -> IntVector:
        return _check_type(self._meta["signal_pdgs"], np.ndarray)

    @property
    def com_energy(self) -> Tuple[float, str]:
        return (_check_type(self._meta['com_e'], float),
                _check_type(self._meta['e_unit'], str))

    @deprecated
    def get_custom_meta(self, name: str) -> Any:
        """Returns user-defined metadata stored under the give name."""
        return self._meta[name]


@_export
class HdfReader(ReaderBase):
    """Create a new heparchy hdf5 file object with read access.
    Processes stored within are accessed via string subscripting.

    :group: hepread

    Parameters
    ----------
    path : pathlib.Path
        Location of the file to be read.

    Examples
    --------
    Opening a file using the context manager:

        >>> import heparchy
        >>> with heparchy.read.HdfReader("showers.hdf5") as hep_file:
        >>>     ...

    """
    def __init__(self, path: Path) -> None:
        self.path = path

    def __enter__(self) -> HdfReader:
        self._buffer = h5py.File(self.path, 'r')
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self._buffer.close()

    def __getitem__(self, key: str) -> HdfProcessReader:
        return HdfProcessReader(self, key=key)

    @deprecated
    def read_process(self, name: str) -> HdfProcessReader:
        return self[name]



from typing import Any, TypeVar

import numpy as np
import numpy.typing as npt


IntVector = npt.NDArray[np.int32]
HalfIntVector = npt.NDArray[np.int16]
DoubleVector = npt.NDArray[np.float64]
BoolVector = npt.NDArray[np.bool_]
AnyVector = npt.NDArray[Any]  # until numpy sorts out structured typing
DataType = TypeVar("DataType")
Self = TypeVar("Self")



"""
`heparchy.write`
================

Interfaces to write to standard particle physics formats,
and the heparchy formats.

Notes
-----
Currently only implemented for hdf5.
"""
from . import hdf
from .hdf import *


__all__ = hdf.__all__.copy()



"""
`heparchy`
==========

The hierarchically formatted high energy physics IO library.
"""
import warnings

from heparchy import read, write
from heparchy._version import __version__, __version_tuple__


__all__ = ["read", "write", "__version__", "__version_tuple__"]
warnings.filterwarnings("once", category=DeprecationWarning)
